"""Interact with LabVIEW VIs from Python"""
# pylint:disable=no-name-in-module
from pathlib import Path
from tempfile import TemporaryDirectory
from enum import IntEnum
import asyncio
from datetime import timezone
import itertools
import warnings
import struct
import pythoncom
from pythoncom import VT_ARRAY, VT_BYREF, VT_I2, VT_UI4, VT_UI1
import pywintypes
import win32com.client
from win32com.client import VARIANT
import numpy as np
from autolv.vistrings import parse_vistrings
from autolv.datatypes import make_control, DataFlow, valididentifier, TimeStamp


class ExecState(IntEnum):
    """ActiveX ExecStateEnum"""

    eBad = 0
    eIdle = 1
    eRunTopLevel = 2
    eRunning = 3


class FPState(IntEnum):
    """Front panel state"""

    Invalid = 0
    Standard = 1
    Closed = 2
    Hidden = 3
    Minimized = 4
    Maximized = 5


class VI:
    """ActiveX connection to a VI

    Parameters
    ----------
    viref : win32com.client.CDispatch
    """

    def __init__(self, viref: win32com.client.CDispatch):
        self._vi = viref
        self.name = self._vi.Name
        methods = [
            "ExportVIStrings",
            "SetControlValue",
            "GetControlValue",
            "Call",
            "Run",
            "Abort",
            "OpenFrontPanel",
            "GetPanelImage",
            "CloseFrontPanel",
            "ReinitializeAllToDefault",
        ]
        for method in methods:
            self._vi._FlagAsMethod(method)
        self.name = self._vi.Name
        with TemporaryDirectory() as tmpdir:
            file = Path(tmpdir).joinpath(self.name)
            self._vi.ExportVIStrings(str(file.absolute()))
            with open(file, "r") as f:
                vistr = f.read()
        self._ctrls = {k: make_control(**v) for k, v in parse_vistrings(vistr).items()}
        for ctrl in self._ctrls.values():
            value = self._vi.GetControlValue(ctrl.name)
            if isinstance(ctrl, TimeStamp):
                value = value.replace(tzinfo=None)
            ctrl.value = value

    def __getitem__(self, item):
        return self._ctrls[item]

    def __dir__(self):
        ctrls = []
        for ctrl in self._ctrls:
            if valididentifier(ctrl):
                ctrls.append(ctrl)
            else:
                ctrls.append(f"['{ctrl}']")
        attrs = super().__dir__()
        attrs = [a for a in attrs if not a.startswith("_")]
        return ctrls + attrs

    def __getattr__(self, item):
        try:
            value = self._ctrls[item]
        except KeyError:
            value = super().__getattribute__(item)
        return value

    def __setitem__(self, item, value):
        self.__setattr__(item, value)

    def __setattr__(self, item, value):
        if "_ctrls" in self.__dict__ and item in self._ctrls:
            self._ctrls[item].value = value
        else:
            super().__setattr__(item, value)

    async def _run(self):
        try:
            self._vi.Run(True)
            while True:
                await asyncio.sleep(0.1)
                if ExecState(self._vi.ExecState) == ExecState.eIdle:
                    break
        except asyncio.CancelledError:
            self._vi.Abort()
        else:
            for ctrl in self._ctrls.values():
                if ctrl._dataflow in [DataFlow.INDICATOR, DataFlow.UNKNOWN]:
                    value = self._vi.GetControlValue(ctrl.name)
                    if isinstance(ctrl, TimeStamp):
                        value = value.replace(tzinfo=None)
                    ctrl.value = value

    async def _spinner(self):
        spinner = ["-", "\\", "|", "/", "-", "|"]
        glyph = itertools.cycle(spinner)
        try:
            while True:
                print(next(glyph), end="\r")
                await asyncio.sleep(0.2)
                if ExecState(self._vi.ExecState) == ExecState.eIdle:
                    break
        except asyncio.CancelledError:
            pass
        print(" ", end="\r")

    async def _task(self):
        try:
            await asyncio.gather(self._run(), self._spinner())
        except asyncio.CancelledError:
            pass

    def run(self, timeout: float = None) -> None:
        """Run the VI and retrieve front panel indicator values

        Parameters
        ----------
        timeout: float
            Allows the VI to run up to timeout seconds. A timeout forces
            the VI to abort. Default of None disables the timeout. This
            function is synchronous and blocks until the VI stops.

        Notes
        -----
        Jupyter has a running event loop and we have to await this function from
        within the notebook:

        In [1]: import autolv
        In [2]: lv = autolv.App()
        In [3]: vi = lv.get_VI(<file>)
        In [4]: await vi.run()

        """
        for ctrl in self._ctrls.values():
            if ctrl._dataflow in [DataFlow.CONTROL, DataFlow.UNKNOWN]:
                value = ctrl.value
                # special case TimeStamp due to timezone issue - see TimeStamp definition
                if isinstance(ctrl, TimeStamp):
                    value = value.replace(tzinfo=timezone.utc)
                self._vi.SetControlValue(ctrl.name, value)
        task = asyncio.wait_for(self._task(), timeout=timeout)
        try:
            asyncio.run(task)
        except asyncio.TimeoutError as exc:
            raise asyncio.TimeoutError(f"{self.name} did not complete") from exc
        except RuntimeError:
            # there's a running event loop already, e.g. Jupyter
            return task
        return None

    @property
    def frontpanel_state(self) -> FPState:
        """Set the front panel state

        Parameters
        ----------
        value : FPState
        """
        return FPState(self._vi.FPState)

    @frontpanel_state.setter
    def frontpanel_state(self, value: FPState) -> None:
        current_state = self._vi.FPState
        if current_state == FPState.Closed:
            self._vi.OpenFrontPanel(True, value)
        elif current_state != FPState.Closed and value == FPState.Closed:
            self._vi.CloseFrontPanel()
        else:
            self._vi.FPState = value

    def get_frontpanel_image(self) -> np.ndarray:
        """Get the front panel image as 8-bit RBGA formatted array

        Returns
        -------
        image : numpy.ndarray
        """

        def rgba(bmp, clrs):
            img = []
            for row in bmp:
                pixels = []
                for pixel in row:
                    color = int(clrs[pixel])  # 24-bit color value
                    rgb = struct.unpack(">BBB", color.to_bytes(3, "big"))
                    pixel = list(rgb) + [255]
                    pixels.append(pixel)
                img.append(pixels)
            return np.array(img)

        # GetPanelImage() returns 8-bit 256 color image regardless of color_depth
        color_depth = 256
        fpsize = self._vi.FPSize
        bitmap = VARIANT(VT_ARRAY | VT_BYREF | VT_UI1, [0] * fpsize)
        colors = VARIANT(VT_ARRAY | VT_BYREF | VT_UI4, [0] * color_depth)
        bounds = VARIANT(VT_ARRAY | VT_BYREF | VT_I2, [0] * 4)
        self._vi.GetPanelImage(True, color_depth, bitmap, colors, bounds)
        # bounds appears to be (0, 0, cols, rows)
        cols, rows = np.array(bounds.value)[2:4]
        bitmap = np.array(bitmap.value).reshape((rows, cols))
        colors = np.array(colors.value)
        return rgba(bitmap, colors)

    def reinitialize_values(self):
        """Reinitialize all controls to default values

        Same as Edit->Reinitialize Values to Default
        """
        self._vi.ReinitializeAllToDefault()
        for ctrl in self._ctrls.values():
            value = self._vi.GetControlValue(ctrl.name)
            if isinstance(ctrl, TimeStamp):
                value = value.replace(tzinfo=None)
            ctrl.value = value


# pylint:disable=attribute-defined-outside-init
class App:
    """ActiveX connection to LabVIEW application"""

    def __init__(self):
        self.__enter__()

    @property
    def version(self) -> str:
        """LabVIEW version"""
        return self._lv.Version

    def get_VI(self, vi_name: str) -> VI:
        """Instantiate a VI object

        Parameters
        ----------
        vi_name : str or path-like
        """
        warnings.warn("get_VI will be removed in v0.3.0", FutureWarning, stacklevel=2)
        return self.open(vi_name)

    def open(self, file_name: str) -> VI:
        """Open a LabVIEW VI

        Parameters
        ----------
        file_name : str or path-like
            VI to open
        """
        file = Path(file_name)
        if file.suffix == ".vi":
            viref = self._lv.GetVIReference(str(file.absolute()))
            return VI(viref)
        raise NotImplementedError(f"'*{file.suffix}' not supported")

    def explain_error(self, code: int) -> str:
        """Explain Error

        Parameters
        ----------
        code : int
            error code from error cluster

        Returns
        -------
        explanation : str
            possible reason as found in LabVIEW's error database
        """
        self._errvi.SetControlValue("Error Code", code)
        self._errvi.Run()
        return self._errvi.GetControlValue("Error Text")

    def close(self):
        """Exit LabVIEW (File -> Exit) if the application was not already running.

        Notes
        -----
        LabVIEW will remain open if the application was already running prior to
        instantiating App. If successfully closed, the App object will be invalid.
        """
        try:
            self._lv.Quit()
        except AttributeError:
            # AttributeError: _lv doesn't yet exist - user needs to __enter__()
            pass
        else:
            # workaround to determine if LabVIEW still running
            try:
                self._lv.Version
            # pylint:disable=no-member
            except (pywintypes.com_error, AttributeError):
                # LabVIEW is actually closed at this point
                self._lv = None
                self._errvi = None

    def __enter__(self):
        if not hasattr(self, "_lv") or self._lv is None:
            # CoInitialize to support running under various processes
            # e.g. rpyc server
            pythoncom.CoInitialize()  # pylint:disable=no-member
            self._lv = win32com.client.Dispatch("LabVIEW.Application")
            methods = ["Quit"]
            for method in methods:
                self._lv._FlagAsMethod(method)
            errvipath = str(
                Path(self._lv.ApplicationDirectory)
                .joinpath(r"vi.lib\Utility\error.llb\Error Code Database.vi")
                .absolute()
            )
            self._errvi = self._lv.GetVIReference(errvipath)
            self._errvi._FlagAsMethod("Run")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __repr__(self):
        if self._lv is None:
            return "<LabVIEW not running>"
        return f"<LabVIEW {self.version}>"



"""Parse Exported VI Strings"""
import re
import io
import xml.etree.ElementTree as ET


def _quote(match: re.Match) -> str:
    """Add quotes to '=attribute'"""
    attr, end = match.groups()
    return f'="{attr}"{end}'


def _addquotes(vistr: str) -> str:
    """Add quotes to '=attribute' attributes"""
    vistr = re.subn(r'=(?!")(.*?)([\s>])', _quote, vistr)[0]
    return vistr


def _close_elements(vistr: str) -> str:
    """Close elements NO_TITLE CRLF LF"""
    # pylint: disable=cell-var-from-loop
    for element in [
        "NO_TITLE",
        "FONT",
        "LF",
        "CRLF",
        "SAME_AS_LABEL",
        "append",
    ]:
        vistr = re.subn(
            f"<{element}.*?>", lambda m: m.group() + f"</{element}>", vistr
        )[0]
    return vistr


def _de_embed_elements(vistr: str) -> str:
    """De-embed elements <<...>>"""
    # pylint: disable=cell-var-from-loop
    for element in [
        "B",
        "append",
    ]:
        vistr = re.subn(f"<(</?{element}>)>", lambda m: m.groups()[0], vistr)[0]
    vistr = re.subn("<<([0-9]+)>>", lambda m: "__" + m.groups()[0] + "__", vistr)[0]
    return vistr


class ClosedInterval:
    """closed interval 'span' from re.match"""

    def __init__(self, span):
        self.start, self.stop = span
        self.stop = self.stop - 1

    def __contains__(self, value):
        return self.start <= value <= self.stop

    def __repr__(self):
        return f"<[{self.start}, {self.stop}]>"


# XML escapes ' " & < > as &...;
# LabVIEW chose not to follow the standard
#   ' comes across as '
#   " -> ""
#   & -> &
#   < -> <<
#   > -> >>
PREDEFINEDS = {
    "&(?!lt;|gt;)": ("amp", 1),
    '""': ("quot", 2),
}


def _predefinedentities(vistr: str) -> str:
    """convert predefined entities"""
    # asumming embedded tags have been deembedded
    vistr = re.subn("<<", "&lt;", vistr)[0]
    vistr = re.subn(">>(?!>)", "&gt;", vistr)[0]
    # exclude altering text within tags
    tag_spans = []
    for tag in re.finditer("<[^<]/?.*?[^>]>", vistr):
        tag_spans.append(ClosedInterval(tag.span()))
    r_vistr = vistr
    for pd, (sub, l) in PREDEFINEDS.items():
        for matched_ch in re.finditer(pd, vistr):
            loc = matched_ch.start()
            if not any(map(lambda span, l=loc: l in span, tag_spans)):
                r_vistr = r_vistr[0:loc] + f"&{sub};" + r_vistr[loc + l :]
    return r_vistr


def _styledtext(vistr):
    """convert styled text to CDATA"""
    vistr = re.subn("<B>(.*?)</B>", lambda m: f"<![CDATA[{m.groups()[0]}]]>", vistr)[0]
    return vistr


def _removeparttext(vistr):
    """Remove <PART ... type="Text">"""
    vistr = re.subn(
        '(<PART .*?type="Text">)(.*?)(</PART>)',
        lambda m: f"{m.groups()[0]}{m.groups()[2]}",
        vistr,
    )[0]
    return vistr


def _vistr2xml(vistr: str) -> str:
    """Fix exported VI strings to compliant XML"""
    vistr = _removeparttext(vistr)
    vistr = _addquotes(vistr)
    vistr = _de_embed_elements(vistr)
    vistr = _close_elements(vistr)
    vistr = _predefinedentities(vistr)
    vistr = _styledtext(vistr)
    return vistr


def _recurse_ctrls(element: ET.Element) -> dict:
    controls = {}
    element = element.find("CONTENT")
    if element is None:
        return controls
    for control in element.iterfind("CONTROL"):
        attrs = {}
        attrs.update(control.attrib)
        desc = control.find("DESC").text
        tip = control.find("TIP").text
        attrs.update({"description": desc, "tip": tip})
        for part in control.find("PARTS"):
            parttype = part.attrib["type"].lower().replace(" ", "")
            try:
                text = part.find("LABEL").find("STEXT").text
            except AttributeError:
                pass
            else:
                attrs.update({parttype: text})
        controls[attrs["name"]] = attrs
        if attrs["type"] == "Cluster":
            controls[attrs["name"]]["ctrls"] = _recurse_ctrls(control)
    return controls


def parse_vistrings(vistr: str) -> dict:
    """Parse exported VI strings pseudo-XML to extract controls"""
    vixml = io.StringIO(_vistr2xml(vistr))
    tree = ET.parse(vixml)
    root = tree.getroot()
    return _recurse_ctrls(root)



"""A Python emulation of LabVIEW's Cluster"""
from abc import ABC, abstractmethod
from collections.abc import Sequence
from numbers import Number
from enum import IntEnum
import re
import pathlib
from datetime import datetime, timezone
import numpy as np

READONLY_ATTRIBUTES = [
    "ID",
    "type",
    "name",
    "description",
    "tip",
    "caption",
    "unitlabel",
]


class DataFlow(IntEnum):
    """Data flow direction"""

    CONTROL = 1
    IN = 1
    INDICATOR = 2
    OUT = 2
    UNKNOWN = 3


def valididentifier(item: str) -> bool:
    """Test 'item' for valid Python identifier"""
    return bool(re.match(r"^[a-zA-Z][\w]+$", item))


class LV_Control(ABC):
    """Abstract base class for LabVIEW control types

    Parameters
    ----------
    kwargs
    """

    # LV control's Label is 'name' and must be unique
    def __init__(self, **kwargs):
        self.name = kwargs.pop("name")
        for attr in READONLY_ATTRIBUTES:
            try:
                setattr(self, attr, kwargs.pop(attr))
            except KeyError:
                pass
        # Exported VI strings does not indicate dataflow direction
        self._dataflow = DataFlow.UNKNOWN

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    def __setattr__(self, item, value):
        if item in self.__dict__ and item in READONLY_ATTRIBUTES:
            raise AttributeError(f"can't set {item}")
        super().__setattr__(item, value)

    def __hash__(self):
        return hash(self.name)

    def __dir__(self):
        attrs = [a for a in self.__dict__ if not a.startswith("_")]
        return attrs

    def set_dataflow(self, direction: str) -> None:
        """Set the control's dataflow direction

        Parameters
        ----------
        direction : str {'control', 'in', 'indicator', 'out'}
        """
        self._dataflow = DataFlow[direction.upper()]

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        tests = []
        for selfattr, otherattr in zip(self.__dict__.keys(), other.__dict__.keys()):
            tests.append(getattr(self, selfattr) == getattr(other, otherattr))
        return all(tests)


class Numeric(LV_Control):
    """Numeric"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.pop("value", 0.0)

    def __setattr__(self, item, value):
        if item == "value":
            if not isinstance(value, Number):
                raise TypeError(f"'{value}' not a number")
        super().__setattr__(item, value)

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"


class Boolean(LV_Control):
    """Boolean"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.pop("value", False)

    def __setattr__(self, item, value):
        if item == "value":
            if not isinstance(value, bool):
                raise TypeError(f"'{value}' not a boolean")
        super().__setattr__(item, value)

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"

    def __bool__(self):
        return self.value


class Array(LV_Control):
    """Array"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.pop("value", np.array([]))

    def __setattr__(self, item, value):
        if item == "value":
            if isinstance(value, str) or not hasattr(value, "__iter__"):
                raise TypeError(f"'{value}' not array like")
            value = np.array(value)
        super().__setattr__(item, value)

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"


class Cluster(LV_Control, Sequence):
    """Cluster - A Python emulation of LabVIEW's Cluster

    A cluster in LabVIEW appears as a key-value mapping but is actually a C-struct
    where the variables (controls) have a set order and are arranged in a contiguous
    block of bytes. The ActiveX 'SetControlValue' expects a Sequence such as a 'list'
    or 'tuple' that follows the same ordering as in LabVIEW.

    In Python, a cluster is more naturally represented as a 'dict'. This 'Cluster'
    combines features of lists and dicts to facilitate communication with LabVIEW
    clusters.

    Parameters
    ----------
    kwargs
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._ctrls = {}
        for name, attrs in kwargs.get("ctrls", {}).items():
            self._ctrls[name] = make_control(**attrs)

    def __getitem__(self, control):
        if isinstance(control, int):
            control = list(self._ctrls)[control]
        return self._ctrls[control]

    def __setitem__(self, control, value):
        if isinstance(control, int):
            control = list(self._ctrls)[control]
        self._ctrls[control].value = value
        values = [ctrl.value for ctrl in self._ctrls]
        self._setfp(self.name, values)

    def __setattr__(self, item, value):
        if "_ctrls" in self.__dict__ and item in self._ctrls:
            self._ctrls[item].value = value
        elif item == "value":
            for c, v in zip(self._ctrls, value):
                self._ctrls[c].value = v
        else:
            super().__setattr__(item, value)

    def __getattr__(self, item):
        if item in self._ctrls:
            value = self._ctrls[item]
        elif item == "value":
            value = [c.value for c in self._ctrls.values()]
        else:
            value = super().__getattribute__(item)
        return value

    def __iter__(self):
        for _, ctrl in self._ctrls.items():
            yield ctrl

    def __len__(self):
        return len(self._ctrls)

    # pylint: disable=arguments-differ
    # Sequence.index start and stop parameters are recommended, not required
    def index(self, control):
        """Return the index of 'control'"""
        for idx, k in enumerate(self._ctrls):
            if k == control:
                return idx
        raise ValueError(f"{control} not in Cluster")

    def count(self, _):
        raise NotImplementedError

    def as_dict(self):
        """Return controls as a dict"""
        return self._ctrls

    def __repr__(self):
        return f"Cluster({self.as_dict()})"

    def __str__(self):
        return str(self.as_dict())

    def __contains__(self, control):
        return control in self._ctrls

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        tests = []
        for selfctrl, otherctrl in zip(self._ctrls.keys(), other._ctrls.keys()):
            tests.append(selfctrl == otherctrl)
            tests.append(getattr(self, selfctrl) == getattr(other, otherctrl))
        return all(tests)

    def __bool__(self):
        return len(self) > 0

    def update(self, controls: dict):
        """Update from a dict"""
        self._ctrls.update(controls)

    def __dir__(self):
        attrs = super().__dir__()
        attrs = [a for a in attrs if not a.startswith("_")]
        ctrls = []
        for ctrl in self._ctrls:
            if valididentifier(ctrl):
                ctrls.append(ctrl)
            else:
                ctrls.append(f"['{ctrl}']")
        return attrs + ctrls


class String(LV_Control):
    """String"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.pop("value", "")

    def __setattr__(self, item, value):
        if item == "value":
            if not isinstance(value, str):
                raise TypeError(f"'{value}' not a string")
        super().__setattr__(item, value)

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"


class Path(LV_Control):
    """Path"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = pathlib.Path(kwargs.pop("value", ""))

    def __setattr__(self, item, value):
        if item == "value":
            if not isinstance(value, (str, pathlib.Path)):
                raise TypeError(f"'{value}' not a string")
            value = pathlib.Path(value)
        super().__setattr__(item, value)

    def __getattribute__(self, item):
        res = super().__getattribute__(item)
        return str(res) if item == "value" else res

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"


class TimeStamp(LV_Control):
    """Time Stamp"""

    # LV 'Time Stamp' comes across ActiveX as naive but pywin32 treats it as UTC
    # The approach here is to force the value that comes from ActiveX to naive
    LV_EPOCH = datetime(1904, 1, 1, tzinfo=timezone.utc)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.pop("value", self.LV_EPOCH)

    def __setattr__(self, item, value):
        if item == "value":
            if not isinstance(value, datetime):
                raise TypeError(f"'{value}' not a datetime")
        super().__setattr__(item, value)

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"


class Enum(LV_Control):
    """Enum"""

    # LV Enum comes across ActiveX as an integer
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.pop("value", 0)

    def __setattr__(self, item, value):
        if item == "value":
            if not isinstance(value, int):
                raise TypeError(f"'{value}' not an integer")
        super().__setattr__(item, value)

    def __repr__(self):
        return f"{self.value}"

    def __str__(self):
        return f"{self.value}"


class IORefNum(LV_Control):
    """I/O ref num"""

    # some I/O ref nums come across ActiveX as a tuple ('<str>', <int>)
    # but end user only needs to know the <str> e.g. PXI1Slot1
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.pop("value", ("", 0))

    def __setattr__(self, item, value):
        if item == "value":
            if isinstance(value, str):
                value = (value, 0)
            if not isinstance(value, tuple):
                raise TypeError(f"'{value}' not a tuple (<str>, <int>)")
        super().__setattr__(item, value)

    def __repr__(self):
        return f"{self.value[0]}"

    def __str__(self):
        return f"{self.value[0]}"


class IVILogicalName(IORefNum):
    """IVI Logical Name"""


class VISAResourceName(IORefNum):
    """VISA resource name"""


class SharedVariable(IORefNum):
    """Shared Variable"""


LVControl_LU = {
    "Numeric": Numeric,
    "Boolean": Boolean,
    "String": String,
    "Path": Path,
    "Time Stamp": TimeStamp,
    "Waveform Graph": LV_Control,
    "Enum": Enum,
    "IVI Logical Name": IVILogicalName,
    "Slide": Numeric,
    "Array": Array,
    "Cluster": Cluster,
    "Measurement Data": Cluster,
    "Classic DAQmx Physical Channel": String,
    "VISA resource name": VISAResourceName,
    "Classic Shared Variable Control": SharedVariable,
}


def make_control(**attrs: dict) -> LV_Control:
    """Make LV_Control from VI strings attributes"""
    LV_Control_cls = LVControl_LU[attrs.pop("type")]
    return LV_Control_cls(**attrs)
